@@ -1,7 +1,7 @@
 /*
  * ascpu is the CPU statistics monitor utility for X Windows
- * Copyright (c) 1998-99  Albert Dorofeev <Albert@mail.dma.be>
- * For the updates see http://bewoner.dma.be/Albert/linux/
+ * Copyright (c) 1998-2000  Albert Dorofeev <Albert@tigr.net>
+ * For the updates see http://www.tigr.net/
  *
  * This software is distributed under GPL. For details see LICENSE file.
  */
@@ -25,6 +25,12 @@
 #include <sys/dk.h>
 #endif
 
+#ifdef _AIX32   /* AIX > 3.1 */
+#include <nlist.h>
+#include <sys/param.h>
+#include <sys/sysinfo.h>
+#endif /* _AIX32 */
+
 #include <X11/Xlib.h>
 #include <X11/xpm.h>
 #include <X11/Xatom.h>
@@ -81,6 +87,20 @@ static struct nlist	nlst[] = {
 };
 #endif
 
+#ifdef _AIX32   /* AIX > 3.1 */
+/* Kernel memory file */
+#define KMEM_FILE "/dev/kmem"
+/* Descriptor of kernel memory file */
+static int  kmem;
+/* Offset of sysinfo structure in kernel memory file */
+static long sysinfo_offset;
+/* Structure to access kernel symbol table */
+static struct nlist namelist[] = {
+  { {"sysinfo"}, 0, 0, {0}, 0, 0 },
+  { {0},         0, 0, {0}, 0, 0 }
+};
+#endif /* _AIX32 */
+
 /*
  * The information over the CPU load is always kept in 4 variables
  * The order is:
@@ -472,6 +492,11 @@ void read_stat()
 	int i;
 #endif
 
+#ifdef _AIX32   /* AIX > 3.1 */
+        time_t *p;
+        struct sysinfo sysinfo;
+#endif          
+
 	/*
 	 * Save the old values 
 	 */
@@ -520,6 +545,23 @@ void read_stat()
 	}
 
 #endif		
+
+#ifdef _AIX32   /* AIX > 3.1 */
+        if (lseek (kmem, sysinfo_offset, SEEK_SET) != sysinfo_offset) {
+          printf("ascpu : cannot lseek %s\n", KMEM_FILE);
+          ascpu_cleanup();
+        }
+        if (read (kmem, (char *)&sysinfo, sizeof(struct sysinfo))
+            != sizeof(struct sysinfo)) {
+          printf("ascpu : cannot read %s\n", KMEM_FILE);
+          ascpu_cleanup();
+        }
+        p = sysinfo.cpu;
+        fresh.load[0] = p[CPU_USER];
+        fresh.load[1] = p[CPU_WAIT];
+        fresh.load[2] = p[CPU_KERNEL];
+        fresh.load[3] = fresh.load[4] = p[CPU_IDLE];
+#endif /* _AIX32 */
 				    
 #ifdef __linux__
 
@@ -564,6 +606,63 @@ void read_stat()
 }
 
 
+#ifdef FOR_CLICK
+/*
+ * This function executes an external command while 
+ * checking whether we should drop the privileges.
+ *
+ * Since we might need privileges later we fork and
+ * then drop privileges in one of the instances which
+ * will then execute the command and die.
+ *
+ * This fixes the security hole for FreeBSD and AIX
+ * where this program needs privileges to access
+ * the system information.
+ */
+void ExecuteExternal()
+{
+	uid_t ruid, euid;
+	int pid;
+#ifdef DEBUG
+	printf("ascpu: system(%s)\n",Command);
+#endif
+	if( ! Command ) {
+		return;
+	}
+	ruid = getuid();
+	euid = geteuid();
+	if ( ruid == euid ) {
+		system( Command );
+		return;
+	}
+	pid = fork();
+	if ( pid == -1 ) {
+		printf("ascpu : fork() failed (%s), command not executed", 
+				strerror(errno));
+		return;
+	}
+	if ( pid != 0 ) {
+		/* parent process simply waits for the child and continues */
+		if ( waitpid(pid, 0, 0) == -1 ) {
+			printf("ascpu : waitpid() for child failed (%s)", 
+				strerror(errno));
+		}
+		return;
+	}
+	/* 
+	 * child process drops the privileges
+	 * executes the command and dies
+	 */
+	if ( setuid(ruid) ) {
+		printf("ascpu : setuid failed (%s), command not executed",
+				strerror(errno));
+		exit(127);
+	}
+	system( Command );
+	exit(0);
+}
+#endif
+
 /* 
  * This checks for X11 events. We distinguish the following:
 #ifdef FOR_CLICK
@@ -581,12 +680,7 @@ void CheckX11Events()
                 {
 #ifdef FOR_CLICK
 		case ButtonPress:
-			if( Command ) {
-#ifdef DEBUG
-				printf("ascpu: system(%s)\n",Command);
-#endif
-				system( Command );
-			}
+			ExecuteExternal();
 			break;
 #endif
                 case Expose:
@@ -949,6 +1043,21 @@ void ascpu_initialize(
 		exit(1);
 	}
 #endif
+
+#ifdef _AIX32   /* AIX > 3.1 */
+        nlist ("/unix", namelist);
+        if (namelist[0].n_value == 0) {
+                printf("ascpu : cannot get nlist\n");
+                exit(1);
+        }
+        sysinfo_offset = namelist[0].n_value;
+        kmem = open (KMEM_FILE, O_RDONLY);
+        if (kmem < 0) {
+                printf("ascpu : cannot open %s\n", KMEM_FILE);
+                exit(1);
+        }
+#endif /* _AIX32 */
+
 	/*running_counter = state.hist_samples;*/
 	prepare_format();
 	read_stat();
